From stdpp Require Export strings.
From iris.proofmode Require Import tokens.
Set Default Proof Using "Type".

Inductive sel_pat :=
  | SelPure
  | SelPersistent
  | SelSpatial : bool → sel_pat
  | SelName : bool → string → sel_pat.

Fixpoint sel_pat_pure (ps : list sel_pat) : bool :=
  match ps with
  | [] => false
  | SelPure :: ps => true
  | _ :: ps => sel_pat_pure ps
  end.

Module sel_pat.
Fixpoint parse_go (ts : list token) (k : list sel_pat) : option (list sel_pat) :=
  match ts with
  | [] => Some (reverse k)
  | THat :: TName s :: ts => parse_go ts (SelName true s :: k)
  | TName s :: ts => parse_go ts (SelName false s :: k)
  | TPure :: ts => parse_go ts (SelPure :: k)
  | TAlways :: ts => parse_go ts (SelPersistent :: k)
  | THat :: TSep :: ts => parse_go ts (SelSpatial true :: k)
  | TSep :: ts => parse_go ts (SelSpatial false :: k)
  | _ => None
  end.
Definition parse (s : string) : option (list sel_pat) :=
  parse_go (tokenize s) [].

Ltac parse s :=
  lazymatch type of s with
  | list sel_pat => s
  | list string => eval vm_compute in (SelName false <$> s)
  | list (bool * string) => eval vm_compute in (curry SelName <$> s)
  | string =>
     lazymatch eval vm_compute in (parse s) with
     | Some ?pats => pats | _ => fail "invalid sel_pat" s
     end
  end.
End sel_pat.
